import { KeyValuePair, ScanStorageArgs, WriteOps } from "../../storage/types";
import { FilterTupleValuePairByPrefix, RemoveTupleValuePairPrefix, TuplePrefix, ValueForTuple } from "../typeHelpers";
import { ScanArgs, TxId, Unsubscribe } from "../types";
/** The low-level API for implementing new storage layers. */
export declare type AsyncTupleStorageApi = {
    scan: (args?: ScanStorageArgs) => Promise<KeyValuePair[]>;
    commit: (writes: WriteOps) => Promise<void>;
    close: () => Promise<void>;
};
/** Wraps AsyncTupleStorageApi with reactivity and MVCC */
export declare type AsyncTupleDatabaseApi = {
    scan: (args?: ScanStorageArgs, txId?: TxId) => Promise<KeyValuePair[]>;
    commit: (writes: WriteOps, txId?: TxId) => Promise<void>;
    cancel: (txId: string) => Promise<void>;
    subscribe: (args: ScanStorageArgs, callback: AsyncCallback) => Promise<Unsubscribe>;
    close: () => Promise<void>;
};
/** Wraps AsyncTupleDatabaseApi with types, subspaces, transaction objects, and additional read apis.  */
export declare type AsyncTupleDatabaseClientApi<S extends KeyValuePair = KeyValuePair> = {
    commit: (writes: WriteOps<S>, txId?: TxId) => Promise<void>;
    cancel: (txId: string) => Promise<void>;
    scan: <T extends S["key"], P extends TuplePrefix<T>>(args?: ScanArgs<T, P>, txId?: TxId) => Promise<FilterTupleValuePairByPrefix<S, P>[]>;
    subscribe: <T extends S["key"], P extends TuplePrefix<T>>(args: ScanArgs<T, P>, callback: AsyncCallback<FilterTupleValuePairByPrefix<S, P>>) => Promise<Unsubscribe>;
    close: () => Promise<void>;
    get: <T extends S["key"]>(tuple: T, txId?: TxId) => Promise<ValueForTuple<S, T> | undefined>;
    exists: <T extends S["key"]>(tuple: T, txId?: TxId) => Promise<boolean>;
    subspace: <P extends TuplePrefix<S["key"]>>(prefix: P) => AsyncTupleDatabaseClientApi<RemoveTupleValuePairPrefix<S, P>>;
    /** Arguments to transact() are for internal use only. */
    transact: (txId?: TxId, writes?: WriteOps<S>) => AsyncTupleRootTransactionApi<S>;
};
export declare type AsyncTupleRootTransactionApi<S extends KeyValuePair = KeyValuePair> = {
    scan: <T extends S["key"], P extends TuplePrefix<T>>(args?: ScanArgs<T, P>) => Promise<FilterTupleValuePairByPrefix<S, P>[]>;
    get: <T extends S["key"]>(tuple: T) => Promise<ValueForTuple<S, T> | undefined>;
    exists: <T extends S["key"]>(tuple: T) => Promise<boolean>;
    subspace: <P extends TuplePrefix<S["key"]>>(prefix: P) => AsyncTupleTransactionApi<RemoveTupleValuePairPrefix<S, P>>;
    set: <T extends S>(tuple: T["key"], value: T["value"]) => AsyncTupleRootTransactionApi<S>;
    remove: (tuple: S["key"]) => AsyncTupleRootTransactionApi<S>;
    write: (writes: WriteOps<S>) => AsyncTupleRootTransactionApi<S>;
    commit: () => Promise<void>;
    cancel: () => Promise<void>;
    id: TxId;
    writes: Required<WriteOps<S>>;
};
export declare type AsyncTupleTransactionApi<S extends KeyValuePair = KeyValuePair> = {
    scan: <T extends S["key"], P extends TuplePrefix<T>>(args?: ScanArgs<T, P>) => Promise<FilterTupleValuePairByPrefix<S, P>[]>;
    get: <T extends S["key"]>(tuple: T) => Promise<ValueForTuple<S, T> | undefined>;
    exists: <T extends S["key"]>(tuple: T) => Promise<boolean>;
    subspace: <P extends TuplePrefix<S["key"]>>(prefix: P) => AsyncTupleTransactionApi<RemoveTupleValuePairPrefix<S, P>>;
    set: <T extends S>(tuple: T["key"], value: T["value"]) => AsyncTupleTransactionApi<S>;
    remove: (tuple: S["key"]) => AsyncTupleTransactionApi<S>;
    write: (writes: WriteOps<S>) => AsyncTupleTransactionApi<S>;
};
/** Useful for indicating that a function does not commit any writes. */
export declare type ReadOnlyAsyncTupleDatabaseClientApi<S extends KeyValuePair = KeyValuePair> = {
    scan: <T extends S["key"], P extends TuplePrefix<T>>(args?: ScanArgs<T, P>, txId?: TxId) => Promise<FilterTupleValuePairByPrefix<S, P>[]>;
    get: <T extends S["key"]>(tuple: T, txId?: TxId) => Promise<ValueForTuple<S, T> | undefined>;
    exists: <T extends S["key"]>(tuple: T, txId?: TxId) => Promise<boolean>;
    subspace: <P extends TuplePrefix<S["key"]>>(prefix: P) => ReadOnlyAsyncTupleDatabaseClientApi<RemoveTupleValuePairPrefix<S, P>>;
};
export declare type AsyncCallback<S extends KeyValuePair = KeyValuePair> = (writes: WriteOps<S>, txId: TxId) => void | Promise<void>;
//# sourceMappingURL=../../../src/database/async/asyncTypes.d.ts.map